@@ -11,18 +11,49 @@ This module contains math helper functions.
 import logging
 import math
 import os
+import itertools
 
 import numpy
 import scipy.sparse
 import scipy.linalg
+from scipy.linalg.lapack import get_lapack_funcs, find_best_lapack_type
+
+# scipy is not a stable package yet, locations change, so try to work
+# around differences (currently only concerns location of 'triu' in scipy 0.7 vs. 0.8)
+try:
+    from scipy.linalg.basic import triu
+except ImportError:
+    from scipy.linalg.special_matrices import triu
 
 blas = lambda name, ndarray: scipy.linalg.get_blas_funcs((name,), (ndarray,))[0]
 
 
+
 logger = logging.getLogger("matutils")
 logger.setLevel(logging.INFO)
 
 
+def qr_destroy(la):
+    a = numpy.asfortranarray(la[0])
+    del la[0], la # now `a` is the only reference to the input matrix
+    m, n = a.shape
+    # perform q, r = QR(component); code hacked out of scipy.linalg.qr
+    logger.debug("computing QR of %s dense matrix" % str(a.shape))
+    geqrf, = get_lapack_funcs(('geqrf',), (a,))
+    qr, tau, work, info = geqrf(a, lwork = -1, overwrite_a = True)
+    qr, tau, work, info = geqrf(a, lwork = work[0], overwrite_a = True)
+    del a # free up mem
+    assert info >= 0
+    r = triu(qr[:n, :n])
+    if m < n: # rare case, #features < #topics
+        qr = qr[:, :m] # retains fortran order
+    gorgqr, = get_lapack_funcs(('orgqr',), (qr,))
+    q, work, info = gorgqr(qr, tau, lwork = -1, overwrite_a = True)
+    q, work, info = gorgqr(qr, tau, lwork = work[0], overwrite_a = True)
+    assert info >= 0, "qr failed"
+    assert q.flags.f_contiguous
+    return q, r
+
 
 def corpus2csc(corpus, num_terms, dtype=numpy.float64):
     """
@@ -91,7 +122,7 @@ class Dense2Corpus(object):
     No data copy is made (changes to the underlying matrix imply changes in the 
     corpus).
     """
-    def __init__(self, dense, documents_columns = True):
+    def __init__(self, dense, documents_columns=True):
         if documents_columns:
             self.dense = dense.T
         else:
@@ -100,9 +131,31 @@ class Dense2Corpus(object):
     def __iter__(self):
         for doc in self.dense:
             yield full2sparse(doc.flat)
+    
+    def __len__(self):
+        return len(self.dense)
 #endclass DenseCorpus            
 
 
+class Sparse2Corpus(object):
+    """
+    Convert a matrix in scipy.sparse format into a streaming gensim corpus.
+    """
+    def __init__(self, sparse, documents_columns=True):
+        if documents_columns:
+            self.sparse = sparse.tocsc()
+        else:
+            self.sparse = sparse.tocsr().T # make sure shape[1]=number of docs (needed in len())
+    
+    def __iter__(self):
+        for indprev, indnow in itertools.izip(self.sparse.indptr, self.sparse.indptr[1:]):
+            yield zip(self.sparse.indices[indprev:indnow], self.sparse.data[indprev:indnow])
+    
+    def __len__(self):
+        return self.sparse.shape[1]
+#endclass Sparse2Corpus
+
+
 def vecLen(vec):
     if len(vec) == 0:
         return 0.0
@@ -280,7 +333,7 @@ class MmReader(object):
     matrix at once (unlike scipy.io.mmread). This allows us to process corpora 
     which are larger than the available RAM.
     """
-    def __init__(self, input):
+    def __init__(self, input, transposed=True):
         """
         Initialize the matrix reader. 
         
@@ -292,7 +345,7 @@ class MmReader(object):
         `seek(0)` (e.g. gzip.GzipFile, bz2.BZ2File).
         """
         logger.info("initializing corpus reader from %s" % input)
-        self.input = input
+        self.input, self.transposed = input, transposed
         if isinstance(input, basestring):
             input = open(input)
         header = input.next().strip()
@@ -303,6 +356,8 @@ class MmReader(object):
         for lineNo, line in enumerate(input):
             if not line.startswith('%'):
                 self.numDocs, self.numTerms, self.numElements = map(int, line.split())
+                if not self.transposed:
+                    self.numDocs, self.numTerms = self.numTerms, self.numDocs
                 break
         logger.info("accepted corpus with %i documents, %i terms, %i non-zero entries" %
                      (self.numDocs, self.numTerms, self.numElements))
@@ -338,7 +393,10 @@ class MmReader(object):
         prevId = -1
         for line in fin:
             docId, termId, val = line.split()
+            if not self.transposed:
+                termId, docId = docId, termId
             docId, termId, val = int(docId) - 1, int(termId) - 1, float(val) # -1 because matrix market indexes are 1-based => convert to 0-based
+            assert prevId <= docId, "matrix columns must come in ascending order"
             if docId != prevId:
                 # change of document: return the document read so far (its id is prevId)
                 if prevId >= 0:
