@@ -8,17 +8,16 @@
 """
 Module for Latent Semantic Indexing.
 
-It actually contains several algorithms for decomposition of extremely large 
-corpora, a combination of which effectively and transparently allows building LSI
-models for:
+This module actually contains several algorithms for decomposition of large corpora, a 
+combination of which effectively and transparently allows building LSI models for:
 
-* corpora much larger than RAM: only constant memory needed, independent of 
+* corpora much larger than RAM: only constant memory is needed, independent of 
   the corpus size (though still dependent on the feature set size)
-* corpora that are streamed: documents are only accessed sequentially, not 
-  random-accessed
-* corpora that cannot even be temporarily stored, each document can only be 
+* corpora that are streamed: documents are only accessed sequentially, no
+  random-access
+* corpora that cannot be even temporarily stored: each document can only be 
   seen once and must be processed immediately (one-pass algorithm)
-* distributed computing for ultra large corpora, making use of a cluster of
+* distributed computing for very large corpora, making use of a cluster of
   machines
 
 Wall-clock performance on the English Wikipedia (2G corpus positions, 3.2M 
@@ -26,41 +25,34 @@ documents, 100K features, 0.5G non-zero entries in the final TF-IDF matrix),
 requesting the top 400 LSI factors:
 
 
-======================================== ============ ==================
- algorithm                                 serial      distributed    
-======================================== ============ ==================
- one-pass update algo (chunks=factors)     109h        19h            
- one-pass merge algo (chunks=40K docs)     8.5h        2.3h           
- two-pass randomized algo (chunks=40K)     2.5h        N/A [1]_           
-======================================== ============ ==================
+====================================================== ============ ==================
+ algorithm                                             serial       distributed    
+====================================================== ============ ==================
+ one-pass merge algorithm                              5h14m        1h41m           
+ multi-pass stochastic algo (with 2 power iterations)  5h39m        N/A [1]_
+====================================================== ============ ==================
+
 
 *serial* = Core 2 Duo MacBook Pro 2.53Ghz, 4GB RAM, libVec
 
-*distributed* = cluster of six logical nodes on four physical machines, each with 
-dual core Xeon 2.0GHz, 4GB RAM, ATLAS
+*distributed* = cluster of four logical nodes on three physical machines, each 
+with dual core Xeon 2.0GHz, 4GB RAM, ATLAS
 
-.. [1] The two-pass algo could be distributed too, but most time is already spent 
-   reading/decompressing the input from disk, and the extra network traffic due to 
-   data distribution would likely make it actually *slower*.
+.. [1] The stochastic algo could be distributed too, but most time is already spent 
+   reading/decompressing the input from disk in its 4 passes. The extra network 
+   traffic due to data distribution across cluster nodes would likely make it 
+   actually *slower*.
 
 """
 
 
 import logging
 import itertools
+import sys
 
 import numpy
 import scipy.sparse
 from scipy.sparse import sparsetools
-from scipy.linalg.lapack import get_lapack_funcs, find_best_lapack_type
-
-# scipy is not a stable package yet, locations change, so try to work
-# around differences (currently only concerns location of 'triu' in scipy 0.7 vs. 0.8)
-try:
-    from scipy.linalg.basic import triu
-except ImportError:
-    from scipy.linalg.special_matrices import triu
-
 
 from gensim import interfaces, matutils, utils
 
@@ -69,8 +61,12 @@ logger = logging.getLogger('lsimodel')
 logger.setLevel(logging.INFO)
 
 
+# accuracy defaults for the multi-pass stochastic algo
+P2_EXTRA_DIMS = 100 # set to `None` for dynamic P2_EXTRA_DIMS=k
+P2_EXTRA_ITERS = 2
 
-def clipSpectrum(s, k, discard = 0.001):
+
+def clipSpectrum(s, k, discard=0.001):
     """
     Given eigenvalues `s`, return how many factors should be kept to avoid
     storing spurious (tiny, numerically instable) values. 
@@ -90,49 +86,35 @@ def clipSpectrum(s, k, discard = 0.001):
 
 
 class Projection(utils.SaveLoad):
-    def __init__(self, m, k, docs = None, algo = 'onepass', chunks = None):
+    def __init__(self, m, k, docs=None, use_svdlibc=False):
         """
-        Construct the (U, S) projection from a corpus `docs`. 
+        Construct the (U, S) projection from a corpus `docs`. The projection can 
+        be later updated by merging it with another Projection via `self.merge()`. 
         
-        This is the class taking care of 'core math'; interfacing with corpora, 
-        chunking large corpora etc. is done through the LsiModel class.
-        
-        `algo` is currently one of:
-        
-          * 'onepass'; only a single pass over `docs` is needed
-          * 'twopass'; multiple passes over the input allowed => can use a 
-            faster algorithm.
+        This is the class taking care of the 'core math'; interfacing with corpora, 
+        splitting large corpora into chunks and merging them etc. is done through 
+        the LsiModel class.
         """
         self.m, self.k = m, k
         if docs is not None:
-            # base case decomposition: given a job `docs`, compute its decomposition in-core
-            # results of several base case decompositions can be merged via `self.merge()`
-            if algo == 'twopass':
-                self.u, self.s = stochasticSvd(docs, k, chunks = chunks, num_terms = m)
-            elif algo == 'onepass':
-                if not scipy.sparse.issparse(docs):
-                    docs = matutils.corpus2csc(docs, num_terms = m)
-                if docs.shape[1] <= max(k, 100):
-                    # For sufficiently small chunk size, update directly like `svd(now, docs)` 
-                    # instead of `svd(now, svd(docs))`.
-                    # This improves accuracy and is also faster for small chunks, because
-                    # we need to perform one less svd.
-                    # On larger chunks this doesn't work because we quickly run out of memory.
-                    self.u = docs
-                    self.s = None
-                else:
-                    try:
-                        import sparsesvd
-                    except ImportError:
-                        raise ImportError("for LSA, the `sparsesvd` module is needed but not found; run `easy_install sparsesvd`")
-                    logger.info("computing sparse SVD of %s matrix" % str(docs.shape))
-                    ut, s, vt = sparsesvd.sparsesvd(docs, k + 30) # ask for extra factors, because for some reason SVDLIBC sometimes returns fewer factors than requested
-                    u = ut.T
-                    del ut, vt
-                    k = clipSpectrum(s ** 2, self.k)
-                    self.u, self.s = u[:, :k], s[:k]
+            # base case decomposition: given a job `docs`, compute its decomposition,
+            # *in-core*.
+            if not use_svdlibc:
+                u, s = stochasticSvd(docs, k, chunks=sys.maxint, num_terms=m, 
+                    power_iters=P2_EXTRA_ITERS, extra_dims=P2_EXTRA_DIMS)
             else:
-                raise NotImplementedError("unknown decomposition algorithm: '%s'" % algo)
+                try:
+                    import sparsesvd
+                except ImportError:
+                    raise ImportError("for LSA, the `sparsesvd` module is needed but not found; run `easy_install sparsesvd`")
+                logger.info("computing sparse SVD of %s matrix" % str(docs.shape))
+                if not scipy.sparse.issparse(docs):
+                    docs = matutils.corpus2csc(docs)
+                ut, s, vt = sparsesvd.sparsesvd(docs, k + 30) # ask for extra factors, because for some reason SVDLIBC sometimes returns fewer factors than requested
+                u = ut.T
+                del ut, vt
+                k = clipSpectrum(s ** 2, self.k)
+            self.u, self.s = u[:, :k], s[:k]
         else:
             self.u, self.s = None, None
     
@@ -141,7 +123,7 @@ class Projection(utils.SaveLoad):
         return Projection(self.m, self.k)
     
 
-    def merge(self, other, decay = 1.0):
+    def merge(self, other, decay=1.0):
         """
         Merge this Projection with another. 
         
@@ -153,40 +135,14 @@ class Projection(utils.SaveLoad):
             return
         if self.u is None:
             # we are empty => result of merge is the other projection, whatever it is
-            if other.s is None:
-                # other.u contains a direct document chunk, not svd => perform svd
-                docs = other.u
-                assert scipy.sparse.issparse(docs)
-                if self.m * self.k < 10000:
-                    # SVDLIBC gives spurious results for small matrices.. run full
-                    # LAPACK on them instead
-                    logger.info("computing dense SVD of %s matrix" % str(docs.shape))
-                    u, s, vt = numpy.linalg.svd(docs.todense(), full_matrices = False)
-                else:
-                    try:
-                        import sparsesvd
-                    except ImportError:
-                        raise ImportError("for LSA, the `sparsesvd` module is needed but not found; run `easy_install sparsesvd`")
-                    logger.info("computing sparse SVD of %s matrix" % str(docs.shape))
-                    ut, s, vt = sparsesvd.sparsesvd(docs, self.k + 30) # ask for a few extra factors, because for some reason SVDLIBC sometimes returns fewer factors than requested
-                    u = ut.T
-                    del ut
-                del vt
-                k = clipSpectrum(s ** 2, self.k)
-                self.u = u[:, :k].copy('F')
-                self.s = s[:k]
-            else:
-                self.u = other.u.copy('F')
-                self.s = other.s.copy()
+            self.u = other.u.copy('F')
+            self.s = other.s.copy()
             return
         if self.m != other.m:
-            raise ValueError("vector space mismatch: update has %s features, expected %s" %
+            raise ValueError("vector space mismatch: update is using %s features, expected %s" %
                              (other.m, self.m))
         logger.info("merging projections: %s + %s" % (str(self.u.shape), str(other.u.shape)))
         m, n1, n2 = self.u.shape[0], self.u.shape[1], other.u.shape[1]
-        if other.s is None:
-            other.u = other.u.todense()
-            other.s = 1.0 # broadcasting will promote this to eye(n2) where needed
         # TODO Maybe keep the bases as elementary reflectors, without 
         # forming explicit matrices with ORGQR.
         # The only operation we ever need is basis^T*basis ond basis*component.
@@ -201,21 +157,9 @@ class Projection(utils.SaveLoad):
         c = gemm(1.0, self.u, other.u, trans_a = True)
         gemm(-1.0, self.u, c, beta = 1.0, c = other.u, overwrite_c = True)
         
-        # perform q, r = QR(component); code hacked out of scipy.linalg.qr
-        logger.debug("computing QR of %s dense matrix" % str(other.u.shape))
-        geqrf, = get_lapack_funcs(('geqrf',), (other.u,))
-        qr, tau, work, info = geqrf(other.u, lwork = -1, overwrite_a = True)
-        qr, tau, work, info = geqrf(other.u, lwork = work[0], overwrite_a = True)
-        del other.u
-        assert info >= 0
-        r = triu(qr[:n2, :n2])
-        if m < n2: # rare case, #features < #topics
-            qr = qr[:, :m] # retains fortran order
-        gorgqr, = get_lapack_funcs(('orgqr',), (qr,))
-        q, work, info = gorgqr(qr, tau, lwork = -1, overwrite_a = True)
-        q, work, info = gorgqr(qr, tau, lwork = work[0], overwrite_a = True)
-        assert info >= 0, "qr failed"
-        assert q.flags.f_contiguous
+        other.u = [other.u] # do some reference magic and call qr_destroy, to save RAM
+        q, r = matutils.qr_destroy(other.u) # q, r = QR(component)
+        assert not other.u
         
         # find the rotation that diagonalizes r
         k = numpy.bmat([[numpy.diag(decay * self.s), c * other.s], [matutils.pad(numpy.matrix([]).reshape(0, 0), min(m, n2), n1), r * other.s]])
@@ -233,12 +177,12 @@ class Projection(utils.SaveLoad):
             s_k = numpy.sqrt(s_k)
         
         k = clipSpectrum(s_k ** 2, self.k)
-        u_k, s_k = u_k[:, :k], s_k[:k]
+        u1_k, u2_k, s_k = u_k[:n1, :k].copy('F'), u_k[n1:, :k].copy('F'), s_k[:k]
         
-        # update & rotate current basis U
+        # update & rotate current basis U = [U, U']*[U1_k, U2_k]
         logger.debug("updating orthonormal basis U")
-        self.u = gemm(1.0, self.u, u_k[:n1]) # TODO temporarily creates an extra (m,k) dense array in memory. find a way to avoid this!
-        gemm(1.0, q, u_k[n1:], beta = 1.0, c = self.u, overwrite_c = True) # u = [u,u']*u_k
+        self.u = gemm(1.0, self.u, u1_k) # TODO temporarily creates an extra (m,k) dense array in memory. find a way to avoid this!
+        gemm(1.0, q, u2_k, beta = 1.0, c = self.u, overwrite_c = True)
         self.s = s_k
 #        diff = numpy.dot(self.u.T, self.u) - numpy.eye(self.u.shape[1])
 #        logger.info('orth error after=%f' % numpy.sum(diff * diff))
@@ -262,7 +206,8 @@ class LsiModel(interfaces.TransformationABC):
     
     """
     def __init__(self, corpus=None, numTopics=200, id2word=None, chunks=20000, 
-                 decay=1.0, distributed=False, onepass=False):
+                 decay=1.0, distributed=False, onepass=True, 
+                 power_iters=P2_EXTRA_ITERS, extra_samples=P2_EXTRA_DIMS):
         """
         `numTopics` is the number of requested factors (latent dimensions). 
         
@@ -275,10 +220,15 @@ class LsiModel(interfaces.TransformationABC):
         If you specify a `corpus`, it will be used to train the model. See the 
         method `addDocuments` for a description of the `chunks` and `decay` parameters.
         
-        If your document stream is one-pass only (the stream cannot be repeated),
-        turn on `onepass` to force a single pass SVD algorithm (slower).
+        Turn `onepass` off to force a multi-pass stochastic algorithm.
+        
+        `power_iters` and `extra_samples` affect the accuracy of the stochastic
+        multi-pass algorithm, which is used either internally (`onepass=True`) or
+        as the front-end algorithm (`onepass=False`). Increasing the number of 
+        power iterations improves accuracy, but lowers performance. See [2]_ for 
+        some hard numbers.
 
-        Turn on `distributed` to force distributed computing.
+        Turn on `distributed` to enable distributed computing.
         
         Example:
         
@@ -287,6 +237,8 @@ class LsiModel(interfaces.TransformationABC):
         >>> lsi.addDocuments(corpus2) # update LSI on additional documents
         >>> print lsi[doc_tfidf]
         
+        .. [2] http://nlp.fi.muni.cz/~xrehurek/nips/rehurek_nips.pdf
+        
         """
         self.id2word = id2word
         self.numTopics = int(numTopics)
@@ -350,7 +302,7 @@ class LsiModel(interfaces.TransformationABC):
 
         Setting `decay` < 1.0 causes re-orientation towards new data trends in the 
         input document stream, by giving less emphasis to old observations. This allows
-        SVD to gradually "forget" old observations (documents) and give more 
+        LSA to gradually "forget" old observations (documents) and give more 
         preference to new ones.
         """
         logger.info("updating SVD with new documents")
@@ -363,9 +315,12 @@ class LsiModel(interfaces.TransformationABC):
         
         if not scipy.sparse.issparse(corpus):
             if not self.onepass:
-                # we are allowed multiple passes over input => use a faster, randomized two-pass algo
-                update = Projection(self.numTerms, self.numTopics, corpus, algo = 'twopass', chunks = chunks)
-                self.projection.merge(update, decay = decay)
+                # we are allowed multiple passes over the input => use a faster, randomized two-pass algo
+                update = Projection(self.numTerms, self.numTopics, None)
+                update.u, update.s = stochasticSvd(corpus, self.numTopics, 
+                    num_terms=self.numTerms, chunks=chunks,
+                    extra_dims=P2_EXTRA_DIMS, power_iters=P2_EXTRA_ITERS)
+                self.projection.merge(update, decay=decay)
             else:
                 # the one-pass algo
                 chunker = itertools.groupby(enumerate(corpus), key = lambda (docno, doc): docno / chunks)
@@ -373,7 +328,7 @@ class LsiModel(interfaces.TransformationABC):
                 for chunk_no, (key, group) in enumerate(chunker):
                     # construct the job as a sparse matrix, to minimize memory overhead
                     # definitely avoid materializing it as a dense matrix!
-                    job = matutils.corpus2csc((doc for _, doc in group), num_terms = self.numTerms)
+                    job = matutils.corpus2csc((doc for _, doc in group), num_terms=self.numTerms)
                     doc_no += job.shape[1]
                     if self.dispatcher:
                         # distributed version: add this job to the job queue, so workers can work on it
@@ -394,15 +349,15 @@ class LsiModel(interfaces.TransformationABC):
                 if self.dispatcher:
                     logger.info("reached the end of input; now waiting for all remaining jobs to finish")
                     self.projection = self.dispatcher.getstate()
+#            logging.info("top topics after adding %i documents" % doc_no)
+#            self.printDebug(10)
         else:
             assert not self.dispatcher, "must be in serial mode to receive jobs"
             assert self.onepass, "distributed two-pass algo not supported yet"
             update = Projection(self.numTerms, self.numTopics, corpus.tocsc())
-            self.projection.merge(update, decay = decay)
+            self.projection.merge(update, decay=decay)
             logger.info("processed sparse job of %i documents" % (corpus.shape[1]))
 
-        self.printTopics(5)
-
     
     def __str__(self):
         return "LsiModel(numTerms=%s, numTopics=%s, decay=%s, chunks=%s)" % \
@@ -480,7 +435,7 @@ class LsiModel(interfaces.TransformationABC):
 #endclass LsiModel
 
 
-def printDebug(id2token, u, s, topics, numWords = 10, numNeg = None):
+def printDebug(id2token, u, s, topics, numWords=10, numNeg=None):
     if numNeg is None:
         # by default, print half as many salient negative words as positive
         numNeg = numWords / 2
@@ -496,9 +451,9 @@ def printDebug(id2token, u, s, topics, numWords = 10, numNeg = None):
     
     logger.debug("printing %i+%i salient words" % (numWords, numNeg))
     for topic in sorted(result.iterkeys()):
-        weights = sorted(result[topic], reverse = True)
+        weights = sorted(result[topic], key=lambda x: -abs(x[0]))
         _, most = weights[0]
-        if u[most, topic] < 0.0: # the most significant word has negative sign => flip sign of u[most]
+        if u[most, topic] < 0.0: # the most significant word has a negative sign => flip sign of u[most]
             normalize = -1.0
         else:
             normalize = 1.0
@@ -507,13 +462,13 @@ def printDebug(id2token, u, s, topics, numWords = 10, numNeg = None):
         pos, neg = [], []
         for weight, uvecno in weights:
             if normalize * u[uvecno, topic] > 0.0001:
-                pos.append('%s(%.3f)' % (id2token[uvecno], normalize * u[uvecno, topic]))
+                pos.append('%s(%.3f)' % (id2token[uvecno], u[uvecno, topic]))
                 if len(pos) >= numWords:
                     break
         
         for weight, uvecno in weights:
             if normalize * u[uvecno, topic] < -0.0001:
-                neg.append('%s(%.3f)' % (id2token[uvecno], normalize * u[uvecno, topic]))
+                neg.append('%s(%.3f)' % (id2token[uvecno], u[uvecno, topic]))
                 if len(neg) >= numNeg:
                     break
 
@@ -682,86 +637,125 @@ def iterSvd(corpus, numTerms, numFactors, numIter = 200, initRate = None, conver
     return sterm, svals, sdoc.T
 
 
-def stochasticSvd(corpus, rank, num_terms=None, chunks=20000, extra_dims=None, dtype=numpy.float64, eps=1e-6):
+def stochasticSvd(corpus, rank, num_terms, chunks=20000, extra_dims=None, 
+                  power_iters=0, dtype=numpy.float64, eps=1e-6):
     """
-    Return U, S -- the left singular vectors and the singular values of the streamed 
-    input corpus.
+    Return (U, S): the left singular vectors and the singular values of the streamed 
+    input corpus `corpus` [3]_.
     
     This may actually return less than the requested number of top `rank` factors, 
-    in case the input is of lower rank. Also note that the decomposition is unique
-    up the the sign of the left singular vectors (columns of U).
+    in case the input is of lower rank. The `extra_dims` (oversampling) and especially
+    `power_iters` (power iterations) parameters affect accuracy of the decomposition.
     
-    This is a streamed, two-pass algorithm, without power-iterations. In case you can 
-    only afford a single pass over the input corpus, set `onepass=True` in LsiModel and 
-    avoid using this algorithm.
+    This algorithm uses `2+power_iters` passes over the data. In case you can only 
+    afford a single pass over the input corpus, set `onepass=True` in LsiModel 
+    and avoid using this algorithm directly.
 
     The decomposition algorithm is based on 
     **Halko, Martinsson, Tropp. Finding structure with randomness, 2009.**
     
+    .. [3] If `corpus` is a scipy.sparse matrix instead, it is assumed the whole 
+       corpus fits into core memory and a different (more efficient) code path is chosen.
     """
     rank = int(rank)
     if extra_dims is None:
         samples = max(10, 2 * rank) # use more samples than requested factors, to improve accuracy
     else:
         samples = rank + int(extra_dims)
+    logger.info("using %i extra samples and %i power iterations" % (samples - rank, power_iters))
     
-    if num_terms is None:
-        logger.warning("number of terms not provided; will scan the corpus (ONE EXTRA PASS, MAY BE SLOW) to determine it")
-        num_terms = len(utils.dictFromCorpus(corpus))
-        logger.info("found %i terms" % num_terms)
-    else:
-        num_terms = int(num_terms)
+    num_terms = int(num_terms)
     
     eps = max(float(eps), 1e-9) # must ignore near-zero eigenvalues (probably numerical error); the associated eigenvectors are typically unstable/garbage
     
-    # first pass: construct the orthonormal action matrix Q = orth(Y) = orth(A * O)
+    # first phase: construct the orthonormal action matrix Q = orth(Y) = orth((A * A.T)^q * A * O)
     # build Y in blocks of `chunks` documents (much faster than going one-by-one 
     # and more memory friendly than processing all documents at once)
     y = numpy.zeros(dtype = dtype, shape = (num_terms, samples))
-    logger.info("1st pass: constructing %s action matrix" % str(y.shape))
+    logger.info("1st phase: constructing %s action matrix" % str(y.shape))
     
-    chunker = itertools.groupby(enumerate(corpus), key = lambda (docno, doc): docno / chunks)
-    for chunk_no, (key, group) in enumerate(chunker):
-        logger.info('PROGRESS: at document #%i' % (chunk_no * chunks))
-        # construct the chunk as a sparse matrix, to minimize memory overhead
-        # definitely avoid materializing it as a dense (num_terms x chunks) matrix!
-        chunk = matutils.corpus2csc((doc for _, doc in group), num_terms=num_terms, dtype=dtype) # documents = columns of sparse CSC
-        m, n = chunk.shape
-        assert m == num_terms
-        assert n <= chunks # the very last chunk of A may be smaller
-        logger.debug("multiplying chunk * gauss")
-        o = numpy.random.normal(0.0, 1.0, (n, samples)).astype(dtype) # draw a random gaussian matrix
-        sparsetools.csc_matvecs(num_terms, n, samples, chunk.indptr, # y = y + chunk * o
-                                chunk.indices, chunk.data, o.ravel(), y.ravel())
-        del chunk, o
+    if scipy.sparse.issparse(corpus):
+        m, n = corpus.shape
+        assert num_terms == m, "mismatch in number of features: %i in sparse matrix vs. %i parameter" % (m, num_terms)
+        o = numpy.random.normal(0.0, 1.0, (n, samples)).astype(y.dtype) # draw a random gaussian matrix
+        sparsetools.csc_matvecs(m, n, samples, corpus.indptr, corpus.indices, 
+                                corpus.data, o.ravel(), y.ravel()) # y = corpus * o
+        del o
+        y = y.astype(dtype) # TODO unlike numpy, scipy actually makes a copy even when dtype=y.dtype...marginally inefficient
+        logger.debug("running %i power iterations" % power_iters)
+        for power_iter in xrange(power_iters):
+            y = corpus.T * y
+            y = corpus * y
+    else:
+        chunker = itertools.groupby(enumerate(corpus), key = lambda (docno, doc): docno / chunks)
+        num_docs = 0
+        for chunk_no, (key, group) in enumerate(chunker):
+            logger.info('PROGRESS: at document #%i' % (chunk_no * chunks))
+            # construct the chunk as a sparse matrix, to minimize memory overhead
+            # definitely avoid materializing it as a dense (num_terms x chunks) matrix!
+            chunk = matutils.corpus2csc((doc for _, doc in group), num_terms=num_terms, dtype=dtype) # documents = columns of sparse CSC
+            m, n = chunk.shape
+            assert m == num_terms
+            assert n <= chunks # the very last chunk of A is allowed to be smaller in size
+            num_docs += n
+            logger.debug("multiplying chunk * gauss")
+            o = numpy.random.normal(0.0, 1.0, (n, samples)).astype(dtype) # draw a random gaussian matrix
+            sparsetools.csc_matvecs(num_terms, n, samples, chunk.indptr, # y = y + chunk * o
+                                    chunk.indices, chunk.data, o.ravel(), y.ravel())
+            del chunk, o
+        
+        for power_iter in xrange(power_iters):
+            logger.info("running power iteration #%i" % (power_iter + 1))
+            yold = y.copy()
+            y[:] = 0.0
+            chunker = itertools.groupby(enumerate(corpus), key = lambda (docno, doc): docno / chunks)
+            for chunk_no, (key, group) in enumerate(chunker):
+                logger.info('PROGRESS: at document #%i/%i' % (chunk_no * chunks, num_docs))
+                chunk = matutils.corpus2csc((doc for _, doc in group), num_terms=num_terms, dtype=dtype) # documents = columns of sparse CSC
+                tmp = chunk.T * yold
+                tmp = chunk * tmp
+                del chunk
+                y += tmp
+            del yold
     
     logger.info("orthonormalizing %s action matrix" % str(y.shape))
-    q, r = numpy.linalg.qr(y) # orthonormalize the range
-    del y # Y not needed anymore, free up mem
+    y = [y]
+    q, r = matutils.qr_destroy(y) # orthonormalize the range
+    del y
     samples = clipSpectrum(numpy.diag(r), samples, discard = eps)
-    qt = q[:, :samples].T.copy() # discard bogus columns, in case Y was rank-deficient
+    qt = numpy.asfortranarray(q[:, :samples].T) # discard bogus columns, in case Y was rank-deficient
     del q
     
-    # second pass: construct the covariance matrix X = B * B.T, where B = Q.T * A
-    # again, construct X incrementally, in chunks of `chunks` documents from the streaming 
-    # input corpus A, to avoid using O(number of documents) memory
-    x = numpy.zeros(shape = (samples, samples), dtype = dtype)
-    logger.info("2nd pass: constructing %s covariance matrix" % str(x.shape))
-    chunker = itertools.groupby(enumerate(corpus), key = lambda (docno, doc): docno / chunks)
-    for chunk_no, (key, group) in enumerate(chunker):
-        logger.info('PROGRESS: at document #%i' % (chunk_no * chunks))
-        chunk = matutils.corpus2csc((doc for _, doc in group), num_terms=num_terms, dtype=dtype)
-        b = qt * chunk # dense * sparse matrix multiply
-        x += numpy.dot(b, b.T) # TODO should call the BLAS routine SYRK, but there is no SYRK wrapper in scipy :(
-        del chunk, b
-    
-    # now we're ready to compute decomposition of the small matrix X
-    logger.info("computing decomposition of the %s covariance matrix" % str(x.shape))
-    u, s, vt = numpy.linalg.svd(x) # could use linalg.eigh, but who cares... and svd returns the factors already sorted :)
-    keep = clipSpectrum(s, rank, discard = eps)
+    if scipy.sparse.issparse(corpus):
+        b = qt * corpus
+        logger.info("2nd phase: running dense svd on %s matrix" % str(b.shape))
+        u, s, vt = numpy.linalg.svd(b, full_matrices=False)
+        del b, vt
+    else:
+        # second phase: construct the covariance matrix X = B * B.T, where B = Q.T * A
+        # again, construct X incrementally, in chunks of `chunks` documents from the streaming 
+        # input corpus A, to avoid using O(number of documents) memory
+        x = numpy.zeros(shape = (samples, samples), dtype = dtype)
+        logger.info("2nd phase: constructing %s covariance matrix" % str(x.shape))
+        chunker = itertools.groupby(enumerate(corpus), key = lambda (docno, doc): docno / chunks)
+        for chunk_no, (key, group) in enumerate(chunker):
+            logger.info('PROGRESS: at document #%i/%i' % (chunk_no * chunks, num_docs))
+            chunk = matutils.corpus2csc((doc for _, doc in group), num_terms=num_terms, dtype=dtype)
+            b = qt * chunk # dense * sparse matrix multiply
+            x += numpy.dot(b, b.T) # TODO should call the BLAS routine SYRK, but there is no SYRK wrapper in scipy :(
+            del chunk, b
     
+        # now we're ready to compute decomposition of the small matrix X
+        logger.info("running dense decomposition on %s covariance matrix" % str(x.shape))
+        u, s, vt = numpy.linalg.svd(x) # could use linalg.eigh, but who cares... and svd returns the factors already sorted :)
+        s = numpy.sqrt(s) # sqrt to go back from singular values of X to singular values of B = singular values of the corpus
+        
     logger.info("computing the final decomposition")
-    s = numpy.sqrt(s[:keep]) # sqrt to go back from singular values of X to singular values of B = singular values of the corpus
-    u = numpy.dot(qt.T, u[:, :keep]) # go back from left singular vectors of B to left singular vectors of the corpus
-    return u.astype(dtype), s.astype(dtype)
-    
+    keep = clipSpectrum(s**2, rank, discard=eps)
+    u = numpy.asfortranarray(u[:, :keep])
+    s = s[:keep]
+    gemm = matutils.blas('gemm', u)
+    u = gemm(1.0, qt, u, trans_a=True)
+    return u, s
+
+
